
void DisplayCursor()
{
	char CurCursorChar;
	int line_s, line_l;

	if( c_y>=line_count ) c_y=line_count-1;

	line_l=GetLineLength( c_y );
	line_s=GetLineStart( c_y );
	
	if( c_x<line_l ) CurCursorChar=GetChar( line_s+c_x );
		else CurCursorChar='*';

	if( (c_y>=t_y) & (c_y<t_y+w_h-1) & (c_x>=t_x) & (c_x<t_x+w_w-1) )
	{
		PutChar( c_x-t_x+w_x+1, c_y-t_y+w_y+1, CurCursorChar, 9+16*2 ); //0+16*2
	}
}

void DisplayLine(int LineNumber, int x, int y, int max, int firstX)
{
	int spos,cpos,l;
	char c;
	int col;

	if( LineNumber<line_count )
	{
		spos=GetLineStart(LineNumber);	//lines[LineNumber].start;
		l=GetLineLength(LineNumber);	//lines[LineNumber].length;
	} else
	{
		l=-1;	// now cpos is never below l, so c is always ' '
		spos=0;
	}

	if( (Mark_ismarked==1)&(LineNumber==Mark_Start_Line) ) col=8+16*7; else col=7;

	for( cpos=0; cpos<max; cpos++ )
	{
		if( (cpos+firstX)<l ) c=GetChar(spos+firstX+cpos);
			else c=' ';
		PutChar( x+cpos, y, c, col );
	}
}

void LineOut(const char* c, int y, int col, int x, int max)
{
	int i=0;
	char c2;
	for(;;)
	{
		if(*c!=0) { c2=*c; c++; } else c2=' ';
		PutChar( i+x, y, c2, col );
		i++;
		if( i>max ) break;
	}
}

void Display( int RedrawMenu )
{
	int i;

	Window( w_x, w_y, w_w, w_h, 15, file_name, 1 );
	SetScrollBar( w_x, w_y, w_w, w_h, 15, t_y, line_count );

		//1=Schwarz
		//2=gelb
		//4=Weiss

		// Draw Status Bar
	LineOut(Status_text, 43, Status_color, 1, 77 );
	PutChar( 1, 43, 231, Status_color );
	i=((Status_color&0x0F)<<4)+
          ((Status_color&0xF0)>>4);
	PutChar( 78, 43, 232, i );

		// Draw Menu bar, in case some terminal-message overwrote it
	if( RedrawMenu!=0 ) DisplayMainMenu( -1, -1 );

		// Draw Text
	for( i=0; i<w_h-1; i++ ) // for each line...
	{
		DisplayLine( i+t_y, w_x+1, w_y+i+1, w_w-1, t_x );
	}

	DisplayCursor();
}

void InitDisplay()
{
	w_x=0;
	w_y=0;
	w_w=79;
	w_h=43;

	t_x=0;
	t_y=0;

	c_x=0;
	c_y=0;

	Status_text=" Use arrow-keys to navigate, press ESC for options";
	Status_color=status_color_normal;

	MemCopy( 0xB8000, ScreenSaver, 160*56 );

	LineOut("SimpleEdit v0.4", 52, 7+16*8, 10, 30);

	LineOut("", 55, 1+16*15, 1, 70 ); // Delete all old terminal-stuff
}

void RestoreDisplay()
{
	MemCopy( ScreenSaver, 0xB8000, 160*56 );
	// TO DO: Maybe use Sys32 service (restore V2OS look) here??
}


void InsertCharacter(char c)
{
	int pos=GetLineStart( c_y ) + c_x;
	int i;
	int size=1;

	if( (c==13)&(line_count>=MaxLines-2) )
	{
		Status_text=" ERROR: too many lines (maximum: 8192). ";
		Status_color=status_color_error;
		return;
	}

	if( c==13 ) //** NEWLINE
		{ if( text_style==0 ) size=2; else size=1; }
	else	//** Simple Character
		{ if( c_x>GetLineLength( c_y ) ) size=c_x-GetLineLength( c_y )+1; else size=1; }


	while( text_size+SecurityDistance+size > TB_size ) // in case size>1024 
		AllocAdditionalBuffer( 0x8000 ); // == 32K

	text_size+=size;

	if( c!=13 )
	{
		MoveMemory( pos, size, 0 );

		SetChar( pos, c );
		for( i=0; i<size-1; i++ )
			SetChar( pos-i-1,  ' ' );
	
		SetLineLength( c_y, GetLineLength( c_y ) + size );	
		for( i=c_y+1; i<line_count; i++)
			SetLineStart( i, GetLineStart( i ) + size );
	} else //NEWLINE
	{		
		if( c_x>GetLineLength( c_y ) )
		{
			c_x=GetLineLength( c_y );
			pos=GetLineStart( c_y ) + c_x;
		}

		MoveMemory( pos, size, 0 );

		SetChar( pos, 10 );
		if( text_style==0 )
		{
			SetChar( pos, 13 );
			SetChar( pos+1, 10 );
		}

		line_count++;
		for( i=line_count; i>c_y+1; i-- )
		{
			SetLineLength( i, GetLineLength( i-1 ) );
			SetLineStart( i, GetLineStart( i-1 ) + size );			
		}
		SetLineStart( c_y+1, pos+size );
		SetLineLength( c_y+1, GetLineLength( c_y ) - c_x );
		SetLineLength( c_y, c_x );
		c_y++;
		c_x=-1; // The calling function (DisplayLoop) adds 1 to c_x, so we have to set it to -1!!
	}
}

void DeleteCharacter(char c)
{
	int pos, i, j;
	if( c_x>GetLineLength(c_y) ) return; // don't delete 'virtual' characters
	
	if( (c_x==0) ) // delete complete Line
	{
		if( c_y==0 ) return; //cannot delete first line

		i = GetLineStart( c_y-1 ) + GetLineLength( c_y - 1 ); // pos of last character+1
		i = GetLineStart( c_y ) - i; // Length of unseen block, usually 1 or 2

		pos = GetLineStart( c_y ) - i;
		MoveMemory( pos, i, 1 );

		c_y--;
		c_x=GetLineLength( c_y )+1;

		SetLineLength( c_y, GetLineLength( c_y ) + GetLineLength( c_y+1 ) );

		for( j=c_y+1; j<line_count-1; j++)
		{
			SetLineLength(j, GetLineLength( j+1 ) );
			SetLineStart(j, GetLineStart( j+1 ) - i );
		}
		line_count--;
		text_size-=i;

		return;
	}

	pos=GetLineStart( c_y ) + c_x - 1;
	MoveMemory( pos, 1, 1 ); //Move from pos on one character backwards

	SetLineLength( c_y, GetLineLength( c_y )-1 );
	for( i=c_y+1; i<line_count; i++ )
		SetLineStart( i, GetLineStart( i ) - 1 );
	
	text_size--;
}

// KeyDown : Looks up the Scancode-map of the system-keyhandler
char KeyDown(int nr)
{
	char result;

	asm("
		mov %%gs:0x600+74, %%esi
		add %1, %%esi
		mov %%gs:(%%esi), %%al
		mov %%al, %0
		"
		: "=m" (result)
		: "m" (nr)
		: "esi"
	);
	return result;
}

void CheckCursorVisible()
{
	if( c_y>=t_y+w_h-1 ) t_y=c_y-w_h+2;
	if( c_y<t_y ) t_y=c_y;
	if( c_x>=t_x+w_w-1 ) t_x=c_x-w_w+2;
	if( c_x<t_x ) t_x=c_x;
}

char DisplayLoop()
{
	unsigned char c;
	int cursormove;

	int ctrl_down;
	int alt_down;
	int shift_down;

	for(;;)
	{
		Display(1);

		c=getkey();

		cursormove=0;

			// Since V2OS returns the same ASCII-Code for arrow-keys and numbers,
			// we have to check the Scancode-table for which was pressed

		alt_down = KeyDown( 56 );
		ctrl_down = KeyDown( 29 );
		shift_down = 0;
		if( (KeyDown(54)==1)|(KeyDown(42)==1) ) // WHY does only the right shift-key work??
			shift_down = 1;
			
			//--- Some Experimental Code ---//
/*		if( shift_down==1 ) // deactivated. sometimes the scancodes aren't fille correctly...
			{ Mark_Start_Line=c_y; Mark_ismarked=1; }
*/

/*		if( alt_down==1 ) // deactivated. produces very strange results. seems like ALT and CTRL have the same
					// scancode. strange and curious. and somehow frightning...
			return c;
*/
		switch(c)
		{
		case '2' : if( KeyDown(80)==0 ) break;
			if( ctrl_down== 1 ){
				if( t_y<(line_count-1) ) t_y++;
				cursormove=2;
			} else {
				if( c_y<(line_count-1) ) c_y++;
				cursormove=1;
			} break;

		case '8' : if( KeyDown(72)==0 ) break;
			if( ctrl_down== 1 ){
				if(t_y>0) t_y--;
				cursormove=2;
			} else {
				if(c_y>0) c_y--;
				cursormove=1;
			} break;

		case '6' : if( KeyDown(77)==0 ) break;
			if( ctrl_down== 1 ){
				t_x++;
				cursormove=2;
			} else {
				c_x++;
				cursormove=1;
			} break;

		case '4' : if( KeyDown(75)==0 ) break;
			if( ctrl_down== 1 ){
				if( t_x>0 ) t_x--;
				cursormove=2;
			} else {
				if( c_x>0 ) c_x--;
				cursormove=1;
			} break;

		case '9' : if( KeyDown(73)==0 ) break;
			if( ctrl_down== 1 ){
				if( t_y>w_h ) t_y-=w_h-1; else t_y=0;
				cursormove=2;				
			} else {
				if( c_y>w_h ) c_y-=w_h-1; else c_y=0;
				cursormove=1;
			} break;

		case '3' : if( KeyDown(81)==0 ) break;
			if( ctrl_down== 1 ){
				if( t_y<(line_count-w_h) ) t_y+=w_h-1; else t_y=line_count-1;
				cursormove=2;
			} else {
				if( c_y<(line_count-w_h) ) c_y+=w_h-1; else c_y=line_count-1;
				cursormove=1; 
			} break;

		case '7' : if( KeyDown(71)==0 ) break;
			if( ctrl_down==1 ){
				t_y=0;
				cursormove=2;
			} else {
				c_x=0;
				cursormove=1;
			} break;

		case '1' : if( KeyDown(79)==0 ) break;
			if( ctrl_down==1 ) {
				t_y=line_count-1;
				cursormove=2;
			} else {
				c_x=GetLineLength( c_y );
				cursormove=1;
			} break;

		case 27  : getkey(); // BUGFIX: V2OS returns 2 characters, if ESC is pressed. we have to catch the second
			return 0;
		}

		if( cursormove==0 )
		{
			if( (c==13) | ((c>31)&(c<128)) ) // ENTER or any other valid key
				{ InsertCharacter(c); c_x++; cursormove=1; }

			if( c==8 ) //Backspace
				{ DeleteCharacter(c); if( c_x>0 ) c_x--; cursormove=1; }
		}

		if( cursormove==1 )
		{
			CheckCursorVisible();
		}

	}
}

int ProcessLines()
{
	int b_pos=0;
	int isline=1;
	int cur_line=0;
	char c;
	
	//lines[cur_line].start=0;
	SetLineStart( cur_line, 0 );

	text_size=0;
	text_style=1;
	
	for(;;)
	{
		c=GetChar(b_pos);
		if( ((c==10)|(c==0)|(c==13))&(isline==1) ) //EOL
		{
			SetLineLength( cur_line, b_pos-GetLineStart(cur_line) );
			isline=0;
			if( c==0 ) break;
			b_pos++;
			if( (c==13)&(GetChar(b_pos)==10) ) { b_pos++; text_style=0; }
			continue;
		}
		if( (c==10)|(c==13) ) // and isline==0 => a simple new line
		{
			cur_line++;
			SetLineStart( cur_line, b_pos );
			SetLineLength( cur_line, 0 );
			b_pos++;
			if( (c==13)&(GetChar(b_pos)==10) ) { b_pos++; text_style=0; }
			continue;
		}
		if( c==0) break; //isline=0 & EOF
		if( isline==0 )
		{
			cur_line++;
			SetLineStart( cur_line, b_pos );
			isline=1;
		}
		b_pos++;
		if( cur_line>=MaxLines-1 )
		{
			Error( " Sorry, too many lines (maximum: 8191) (press any key to continue)");
			return 1;
		}
	}

	line_count=cur_line+1;
	text_size=b_pos+1;

	term_puts( SCR_DEBUG, "Line Processing complete, Line count = ");
	term_hexprintl( SCR_DEBUG, line_count );
	term_puts( SCR_DEBUG, "\r\nText Size = ");
	term_hexprintl( SCR_DEBUG, text_size );
	term_puts( SCR_DEBUG, "\r\nText Style (0=DOS, 1=UNIX) = ");
	term_hexprintb( SCR_DEBUG, text_style );
	term_puts( SCR_DEBUG,  "\r\n" );

	return 0;
}
